#! /usr/bin/env python

# import standard modules
from optparse import OptionParser
import os, re, sys, time

# import custom modules
from utils.log import GetLogPath, CloseLogFile
from utils.error import MyError
from utils.general import SetupMainClass, TimeSpent, WriteCommand
from utils.messages import ErrMsg, LogMsg, DebugMsg
from utils.files_paths import (FileBoxCls, CheckFilePath, CheckDirPath,
                               GetOutDir, EnsureAbsPath)
from parsers.gene_annotation import (aceview, ensembl, ensg, knownGene,
                                     refGene)

# CONSTANTS
ES_SUCCESS = 0
ES_RUN_ERR = 1
ES_OPT_ERR = 2
ES_PATH_ERR = 3

ANNOT_TYPES = list([
  'aceview',
  'ensembl',
  'knowngene',
  'refseq',
])

class ExonCoordsCreatorCls: #{
  def __init__(self, options): #{
    SetupMainClass(self, options)
  #} end def

  def __del__(self): #{
    CloseLogFile(self)
  #} end def

  def CreateExonCoordsFile(self): #{
    LogMsg(self, "Creating exon coordinates file...")
    start = time.time()
    # get the annotations type
    if (None == self.options.annotations_type): #{
      annots_type = GetAnnotationsType(self.options.input_file)
    else:
      annots_type = self.options.annotations_type
    #} end if
    parse_annot = GetAnnotationsParseFunction(annots_type, self.log_info)
    # open the input and output files
    in_file = FileBoxCls(self.options.input_file, "r",
      "cannot open gene annotations file")
    out_file = FileBoxCls(self.options.output_path, "w",
      "cannot create exon coordinates file")
    # iterate through the gene annotations file
    LogMsg(self, "Parsing gene annotations file...")
    for line in in_file: #{
      DebugMsg(self, line)
      annot = FixAnnotation(parse_annot(line))
      # filter chromosomes, if the option is being used
      if (self.options.filter_chromosomes): #{
        chr_patt = r"\A(chr)?([1-9][0-9]*|[XY]|MT?)\Z"
        if (None == re.search(chr_patt, annot.chrom)): #{
          DebugMsg(self, "Skipping exon in chromosome %s" % annot.chrom)
          continue
        #} end if
      #} end if
      # check whether the gene is non-coding
      if (annot.cdsStart >= annot.cdsEnd): #{
        DebugMsg(self, "Non-coding exon %s %i-%i" %
          (annot.alias, annot.cdsStart, annot.cdsEnd))
        annot.non_coding = True
        annot.alias += ".non_coding"
      #} end if
      # iterate through the exons for the gene
      for (e_start, e_end) in annot.exons: #{
        DebugMsg(self, "Exon start: %i, end: %i" % (e_start, e_end))
        exon_strs = list()
        # if the gene is non-coding, just write the exon
        if (annot.non_coding): #{
          exon_data = list([
            annot.chrom,
            "%i" % e_start,
            "%i" % e_end,
            annot.alias,
          ])
          exon_strs.append(" ".join(exon_data))
          DebugMsg(self, "  non-coding: %s" % " ".join(exon_data))
        # if the exon ends before the CDS start or
        # the exon starts after the CDS end,
        # the full exon is a UTR
        elif (e_end < annot.cdsStart or annot.cdsEnd < e_start):
          exon_data = list([
            annot.chrom,
            "%i" % e_start,
            "%i" % e_end,
            "%s.UTR" % annot.alias,
          ])
          exon_strs.append(" ".join(exon_data))
          DebugMsg(self, "  full UTR: %s" % " ".join(exon_data))
        else:
          # if the exon starts before the CDS start and
          # ends after the CDS start,
          # the first part of the exon is a UTR
          if (e_start < annot.cdsStart): #{
            utr_data = list([
              annot.chrom,
              "%i" % e_start,
              "%i" % (annot.cdsStart-1),
              "%s.UTR" % annot.alias,
            ])
            exon_strs.append(" ".join(utr_data))
            e_start = annot.cdsStart
            DebugMsg(self, "  UTR start: %s\n  New start: %i" %
              (" ".join(utr_data), e_start))
          #} end if
          # if the exon starts before the CDS end and
          # ends after the CDS end,
          # the second part of the exon is a UTR
          if (annot.cdsEnd < e_end): #{
            exon_data = list([
              annot.chrom,
              "%i" % e_start,
              "%i" % annot.cdsEnd,
              annot.alias,
            ])
            exon_strs.append(" ".join(exon_data))
            utr_data = list([
              annot.chrom,
              "%i" % (annot.cdsEnd+1),
              "%i" % e_end,
              "%s.UTR" % annot.alias,
            ])
            exon_strs.append(" ".join(utr_data))
            DebugMsg(self, "  exon start: %s\n  UTR end: %s" %
              (" ".join(exon_data), " ".join(utr_data)))
          # if the exon starts after the CDS start and
          # ends before the CDS end,
          # the full exon is really an exon
          elif (e_start <= e_end):
            exon_data = list([
              annot.chrom,
              "%i" % e_start,
              "%i" % e_end,
              annot.alias,
            ])
            exon_strs.append(" ".join(exon_data))
            DebugMsg(self, "  full exon: %s" % " ".join(exon_data))
          else:
            raise ExonCoordsError("cannot determine exon type: "
              "%s: CDS:%i-%i, Exon:%i-%i" % (annot.alias, annot.cdsStart,
              annot.cdsEnd, e_start, e_end))
          #} end if
        #} end if
        out_file.WriteLine("%s" % "\n".join(exon_strs))
      #} end for
    #} end for
    out_file.Close()
    LogMsg(self, "Time spent creating exon coordinates file: %s" %
      TimeSpent(start))
  #} end def
#} end class

def GetAnnotationsType(paths): #{
  # make sure paths is a list
  if (not isinstance(paths, list)): #{
    if (isinstance(paths, str)): #{
      if ("," in paths): #{
        paths = paths.split(",")
      elif (";" in paths):
        paths = paths.split(";")
      else:
        paths = [paths]
      #} end if
    elif (isinstance(paths, (set,tuple))):
      paths = list(paths)
    elif (isinstance(paths, dict)):
      paths = paths.values()
    elif (not hasattr(paths, '__iter__')):
      paths = [paths]
    else:
      raise ExonCoordsError("cannot create list from paths object in "
         "GetAnnotationsType: %s (%s)" % (str(paths), type(paths).__name__))
    #} end if
  #} end if
  # check the paths the easy way
  for annot_type in ANNOT_TYPES: #{
    for path in paths: #{
      if (annot_type.lower() in path.lower()): #{
        return annot_type
      #} end if
    #} end for
  #} end for
  # check for special file names
  for path in paths: #{
    if ("acembly" in path.lower()): #{
      return "aceview"
    elif ("ensgene" in path.lower()):
      return "ensembl"
    elif ("refgene" in path.lower()):
      return "refseq"
    #} end if
  #} end for
  # if you have not found an annotations type by this point, raise an error
  raise ExonCoordsError("could not determine annotations type")
#} end def

def GetAnnotationsParseFunction(annotations_type, log_info=dict()): #{
  DebugMsg(log_info, "Getting annotations parser...\n"
    "Annotations type: %s" % annotations_type)
  parse_functions = {
    'aceview':   aceview.parse_line,
    'ensembl':   ensembl.parse_line,
    'ensg':      ensg.parse_line,
    'knowngene': knownGene.parse_line,
    'refseq':    refGene.parse_line,
  }
  if (annotations_type.lower() in parse_functions): #{
    return parse_functions[annotations_type.lower()]
  else:
    raise ExonCoordsError("cannot determine correct annotation parser from "
      "annotations type: %s" % annotations_type)
  #} end if
#} end def

def FixAnnotation(annot): #{
  if (None == annot): #{
    return None
  #} end if
  # replace spaces in the alias
  annot.alias = annot.alias.replace(" ", "_")
  # note that start positions need to be adjusted
  # (but exon start positions were adjusted during parsing)
  annot.txStart = int(annot.txStart)+1
  annot.txEnd = int(annot.txEnd)
  annot.cdsStart = int(annot.cdsStart)+1
  annot.cdsEnd = int(annot.cdsEnd)
  annot.exonCount = int(annot.exonCount)
  # assume that the gene is a coding gene
  annot.non_coding = False
  return annot
#} end def

#### EXCEPTION CLASSES ####
class ExonCoordsError(MyError): #{
  pass
#} end class

def SetupOptionsParser(): #{
  description_string = ("Extract exon coordinates from a gene annotation "
    "file and create a file holding the exon coordinates formatted for the "
    "overlap code, with non-coding genes and UTRs marked.")
  args = [ "INPUT_FILE", "OUTPUT_DIR", ]
  usage_string = "%prog " + " ".join(args) + " [ OPTIONS ]"
  parser = OptionParser(description=description_string,
                        version="%prog 0.1.2",
                        usage=usage_string)
  parser.num_args = len(args)
  parser.add_option("--annotations-type",
                    help="The type of annotations file being used for the "
                         "gene names, e.g.: %s. " % ", ".join(ANNOT_TYPES) +
                         "The code will try to automatically determine the "
                         "annotations type if this option is not used.")
  parser.add_option("--filter-chromosomes",
                    action="store_true",
                    help="Only write out exons for genes on standard "
                         "chromosomes: chr<I>, chrX, chrY, chrM (where "
                         "<I> is any integer). [default]")
  parser.add_option("--all-chromosomes",
                    action="store_false", dest="filter_chromosomes",
                    help="Write out exons for genes on all chromosomes.")
  parser.add_option("-f", "--force",
                    action="store_true",
                    help="Force filtering to take place, overwriting the exon "
                         "coordinates file if it already exists.")
  parser.add_option("-d", "--debug",
                    action="store_true",
                    help="Print debug information while the program runs.")
  parser.set_defaults(filter_chromosomes=True,
                      force=False,
                      debug=False)
  return parser
#} end def

def CheckPaths(options): #{
  path_errors = list()
  opts_good = True
  if (None != options.annotations_type and
      options.annotations_type.lower() not in ANNOT_TYPES):
    ErrMsg("\"%s\" is not a valid annotations type. Valid types are: %s" %
            ", ".join(ANNOT_TYPES))
    opts_good = False
  #} end if
  CheckFilePath(options.input_file, "gene annotations", path_errors)
  if (opts_good and 0 == len(path_errors)): #{
    CheckDirPath(options.output_dir, "output", path_errors, create=True)
  #} end if
  if (opts_good and 0 == len(path_errors)): #{
    input_file_name = os.path.basename(options.input_file)
    (input_root, ext) = os.path.splitext(input_file_name)
    output_root = "%s.exons" % input_root
    if (options.filter_chromosomes): #{
      output_root += ".std_chr"
    #} end if
    output_file_name = output_root + ext
    options.output_path = os.path.join(options.output_dir, output_file_name)
    if (options.debug): #{
      ErrMsg("Output path: %s" % options.output_path)
    #} end if
    if (not options.force and os.path.exists(options.output_path)): #{
      path_errors.append("Exon coordinates file already exists, please "
        "remove it or use --force option to overwrite it: \"%s\"." %
        options.output_path)
    #} end if
    # set the log-file name
    options.log_file_name = GetLogPath(options.input_file,
      "exon_coords", options.output_dir)
  #} end if
  if (0 < len(path_errors)): #{
    ErrMsg("Errors in input arguments:\n  %s" % "\n  ".join(path_errors))
  #} end if
  # the paths are good if there are no path errors and no conflicting options
  return (opts_good and 0 == len(path_errors))
#} end def

def Main(): #{
  # get options and arguments
  parser = SetupOptionsParser()
  (options, args) = parser.parse_args()
  # if the right number of args was used
  if (parser.num_args == len(args)): #{
    options.input_file = EnsureAbsPath(args[0])
    options.output_dir = EnsureAbsPath(args[1])
    if (CheckPaths(options)): #{
      try: #{
        exon_coords_creator = ExonCoordsCreatorCls(options)
        WriteCommand(exon_coords_creator, sys.argv)
        exon_coords_creator.CreateExonCoordsFile()
      except (MyError), e:
        ErrMsg("ERROR while creating exon coordinates file:\n  %s" % e)
        return ES_RUN_ERR
      #} end try
    else:
      return ES_PATH_ERR
    #} end if
  else:
    parser.error("you must specify the path to a file containing gene "
                 "annotations (INPUT_FILE); and the path to a directory "
                 "to hold the output (OUTPUT_DIR).")
    return ES_OPT_ERR
  #} end if
  return ES_SUCCESS
#} end def

if __name__ == '__main__': #{
  exit_status = Main()
  if (ES_SUCCESS == exit_status): #{
    print "EXON COORDS SUCCESS"
  else:
    print "EXON COORDS FAIL"
  #} end if
  sys.exit(exit_status)
#} end if
