#! /usr/bin/env python

# import standard modules
import re, time

# import custom modules
from utils.error import MyError
from utils.general import TimeSpent, RemoveExonTypeLabel
from utils.messages import LogMsg, DebugMsg
from utils.files_paths import GetFilePath, FileBoxCls
from base_predictor import BasePredictorCls
from annotation.create_exon_coords_file import (GetAnnotationsType,
                                                GetAnnotationsParseFunction)

# constants
MIN_IMPROVE_FRACTION = 1.5

class FusionPredictorCls(BasePredictorCls): #{
  def __init__(self, options, log_info=dict()): #{
    BasePredictorCls.__init__(self, 'fusion', 'fus', 'fusion',
      options, log_info=log_info)
    dup_gene_path = GetFilePath(options.output_dir, options.lib,
      "%s.dup_genes" % self.ext)
    self.dup_gene_file = FileBoxCls(dup_gene_path, "w",
      "cannot create duplicated gene %s output file" % self.description)
    mix_dir_path = GetFilePath(options.output_dir, options.lib,
      "%s.mix_dirs" % self.ext)
    self.mix_dir_file = FileBoxCls(mix_dir_path, "w",
      "cannot create mixed gene directions %s output file" % self.description)
    # gene_directions[gene_id] = {+,-,*} where "*" = both directions
    self.gene_directions = dict()
    # get gene directions from annotation file
    if (self.options.use_gene_directions): #{
      self.GetGeneDirections()
    #} end if
  #} end def

  def GetGeneDirections(self): #{
    LogMsg(self, "Getting gene directions from annotations file...")
    start_time = time.time()
    DebugMsg(self, "  %s" % self.options.transcript_annotations)
    conflicted_genes = set()
    # get the appropriate parser
    annot_type = GetAnnotationsType(self.options.transcript_annotations)
    parse_annot = GetAnnotationsParseFunction(annot_type, self.log_info)
    # open the annotations file
    annots_file = FileBoxCls(self.options.transcript_annotations, "r",
      "cannot open annotations file")
    # iterate through the annotations file
    for annot_line in annots_file: #{
      annot = parse_annot(annot_line)
      # replace spaces in the name and alias
      annot.name = annot.name.replace(" ", "_")
      annot.alias = annot.alias.replace(" ", "_")
      annot_name = annot.name.lower()
      annot_alias = annot.alias.lower()
      # skip genes with conflicting annotations
      if (annot_name  in conflicted_genes or
          annot_alias in conflicted_genes): #{
        continue
      #} end if
      # ensure that the gene is on a "normal" chromosome
      chr_patt = r"^(chr)?([1-9][0-9]*|[XY]|MT?)$"
      if (None == re.search(chr_patt, annot.chrom)): #{
        continue
      #} end if
      for field in ["name", "alias"]: #{
        id = getattr(annot, field).lower()
        if (id in self.gene_directions): #{
          if (annot.strand != self.gene_directions[id]): #{
            msg = ("conflicting gene annotations: %s: %s, %s" %
              (getattr(annot, field), annot.strand, self.gene_directions[id]))
            # add the current gene to the list of conflicted genes
            conflicted_genes.update([annot_name, annot_alias])
            if (self.options.ignore_conflicts): #{
              # remove it from the directions dictionary
              if (annot_name in self.gene_directions): #{
                del self.gene_directions[annot_name]
              #} end if
              if (annot_alias in self.gene_directions): #{
                del self.gene_directions[annot_alias]
              #} end if
              DebugMsg(self, "warning: %s" % msg)
            elif (self.options.use_conflicts):
              # mark the gene as going both ways
              self.gene_directions[id] = "*"
            else:
              raise FusionPredictorError(msg)
            #} end if
          #} end if
        else:
          self.gene_directions[id] = annot.strand
        #} end if
      #} end for
    #} end for
    annots_file.Close()
    LogMsg(self, "Time spent getting gene directions: %s" %
      TimeSpent(start_time))
  #} end def

  def TestGroup(self, group): #{
    DebugMsg(self, "Group %s. Topologies: %s" %
      (group.id, ",".join(group.topologies)))
    # gapped candidates are never fusions
    if (group.gap): #{
      DebugMsg(self, "gap candidates are never fusions")
      return False
    #} end if
    # interchromosomal candidates are always fusions
    # unless gene direction is being considered
    #if ("interchr" in group.topologies): #{
    #  if (0 == len(self.gene_directions)): #{
    #    DebugMsg(self, "interchromosomal candidate without considering "
    #      "gene directions")
    #    return True
    #  #} end if
    #else:
    # if the candidate is not interchromosomal, check whether
    # the genomic coordinates overlap
    if ("interchr" not in group.topologies): #{
      # if the candidate's genomic coordinates overlap,
      # then it is not a fusion
      if ((group.coord_pair_A.left <= group.coord_pair_B.left and
           group.coord_pair_B.left <  group.coord_pair_A.right) or
          (group.coord_pair_B.left <  group.coord_pair_A.left and
           group.coord_pair_A.left <  group.coord_pair_B.right)):
        DebugMsg(self, "genomic coordinates overlap")
        return False
      #} end if
      DebugMsg(self, "non-overlapping genomic coordinates")
    #} end if
    #return self.FusionGeneCheck(warnings)
    return True
  #} end def

  def TestMember(self, member): #{
    DebugMsg(self, "  Member %s. Topology: %s" %
      (member.candidate_id, member.topology))
    # if either gene set is empty, then the event is not a fusion
    if (not member.OverlapsGene("A") or not member.OverlapsGene("B")): #{
      DebugMsg(self, "  empty gene set")
      return False
    #} end if
    # if the candidate contig is not interchromosomal,
    # check that the gene sets are disjoint
    if ("interchr" != member.topology): #{
      DebugMsg(self, "  Genes: %s" % member.MainGenesString())
      # if the gene sets overlap at all, then the event is not a fusion
      if (member.GeneSetsOverlap()): #{
        DebugMsg(self, "  overlapping gene sets")
        return False
      #} end if
      DebugMsg(self, "  disjoint gene sets")
    #} end if
    # check gene directions and set 5' and 3' ends
    return self.CheckGeneDirections(member)
    #if (None != self.gene_directions): #{
    #  # check gene directions
    #  if (member.CheckGeneDirections(self.gene_dirs)): #{
    #  #} end if
    #} end if
  #} end def

  def CheckGeneDirections(self, member): #{
    DebugMsg(self, "  checking gene directions...")
    if (None == self.gene_directions): #{
      DebugMsg(self, "  No gene directions to check.")
      return True
    #} end if
    orient_A = self.GetOrientation(member.genes_A,
      member.align_info_A.Strand())
    if ("?" == orient_A): #{
      if (member.OverlapsGene("A")):
        DebugMsg(self, "  could not get first gene direction for candidate.\n"
          "  %s" % member.DataString())
      #} end if
      return False
    #} end if
    orient_B = self.GetOrientation(member.genes_B,
      member.align_info_B.Strand())
    if ("?" == orient_B): #{
      if (member.OverlapsGene("B")):
        DebugMsg(self, "  could not get second gene direction for candidate.\n"
          "  %s" % member.DataString())
      #} end if
      return False
    #} end if
    DebugMsg(self, "  Orientation A: %s, B: %s" % (orient_A, orient_B))
    if (("+" == orient_A and "+" == orient_B) or
        ("+" == orient_A and "*" == orient_B) or
        ("*" == orient_A and "+" == orient_B) or
        ("*" == orient_A and "*" == orient_B)): #{
      member.ends = ("5", "3")
    elif (("-" == orient_A and "-" == orient_B) or
        ("-" == orient_A and "*" == orient_B) or
        ("*" == orient_A and "-" == orient_B)):
      member.ends = ("3", "5")
    elif ("+" == orient_A and "-" == orient_B):
      member.ends = ("5", "5")
    elif ("-" == orient_A and "+" == orient_B):
      member.ends = ("3", "3")
    else:
      raise FusionPredictorError("cannot determine 5' and 3' ends of "
        "fusion, with orientations A: %s, B: %s" % (orient_A, orient_B))
    #} end if
    DebugMsg(self, "  Member ends: %s',%s'" % member.ends)
    if ("*" == orient_A or "*" == orient_B or orient_A == orient_B): #{
      DebugMsg(self, "  valid gene directions")
      return True
    #} end if
    DebugMsg(self, "non-continuous gene directions")
    # note: non-continuous gene direction (mixed-direction) events
    # will be stored in a separate file
    return True
    #return False
  #} end def

  def GetOrientation(self, gene_list, align_strand): #{
    # use the strand of the first gene found from the annotations file
    gene_strand = None
    for gene in gene_list: #{
      gene_id = RemoveExonTypeLabel(gene).lower()
      if (gene_id in self.gene_directions): #{
        gene_strand = self.gene_directions[gene_id]
        break
      #} end if
    #} end for
    # if the gene strand could not be determined
    if (None == gene_strand): #{
      return "?"
    # if the gene goes both ways
    elif ("*" == gene_strand): #{
      return "*"
    # if the gene is on the same strand as the alignment
    elif (gene_strand == align_strand): #{
      return "+"
    # if the gene is on the opposite strand as the alignment
    else:
      return "-"
    #} end if
  #} end def

  def OutputEvent(self, event): #{
    if (self.write_special): #{
      for member in event.members: #{
        # if the member has mixed gene directions
        if (hasattr(member, "ends") and None != member.ends and
            (("5","5") == member.ends or ("3","3") == member.ends)): #{
          DebugMsg(self, "Event with mixed gene directions")
          # write the event to the mixed-directions output file
          self.mix_dir_file.Write(event.FullDataString())
          return
        #} end if
        # if this is an interchromosomal fusion, check whether it
        # involves copies of the same gene on different chromosomes
        if ("interchr" == member.topology and member.GeneSetsOverlap()): #{
          DebugMsg(self, "Interchromosomal fusion involving duplicated gene")
          # write the event to the duplicated-gene output file
          self.dup_gene_file.Write(event.FullDataString())
          return
        #} end if
      #} end for
    #} end if
    BasePredictorCls.OutputEvent(self, event)
  #} end def

  #def ReprocessMember(self, member, contig_alignment, contig_seq=None): #{
  def ReprocessMember(self, member, realigned_contig): #{
    contig_alignment = realigned_contig.best_align
    if (contig_alignment.perfect): #{
      DebugMsg(self, "  False positive: perfect alignment (%i)" %
        contig_alignment.match)
      return False
    #} end if
    new_span = (contig_alignment.ctg_end - contig_alignment.ctg_start) + 1
    min_span = int(max(member.align_info_A.ContigSpan(),
      member.align_info_B.ContigSpan()) * MIN_IMPROVE_FRACTION)
    DebugMsg(self, "  New span: %i, Min: %i" % (new_span, min_span))
    # if the new alignment is significantly longer (in the contig) than
    # the split contig-to-genome alignments, then the contig is probably
    # not really a fusion
    #if (MIN_IMPROVEMENT < (new_span - member.align_info_A.ContigSpan()) and
    #    MIN_IMPROVEMENT < (new_span - member.align_info_B.ContigSpan())): #{
    if (min_span < new_span): #{
      DebugMsg(self, "  False positive: better alignment")
      return False
    #} end if
    region_pair = realigned_contig.region_pairs[member.IDString()]
    if (region_pair.GeneSetsOverlap()): #{
      gene_str_A = "A (%i-%i): %s" % (region_pair.starts["A"],
        region_pair.ends["A"], ",".join(sorted(region_pair.gene_sets["A"])))
      gene_str_B = "B (%i-%i): %s" % (region_pair.starts["B"],
        region_pair.ends["B"], ",".join(sorted(region_pair.gene_sets["B"])))
      DebugMsg(self, "\n    ".join(["  False positive: gene sets overlap",
        gene_str_A, gene_str_B]))
      return False
    #} end if
    DebugMsg(self, "  Good fusion prediction")
    return True
  #} end def
#} end class

#### EXCEPTION CLASSES ####
class FusionPredictorError(MyError): #{
  pass
#} end class
