#! /usr/bin/env python

# TODO
# deal with always excluding non_coding genes when getting ITDs

# import standard modules

# import custom modules
from utils.error import MyError
from utils.general import MaxMatchLength, IsHomopolymerSequence
from utils.messages import LogMsg, DebugMsg
from utils.files_paths import GetFilePath, FileBoxCls
from base_predictor import BasePredictorCls

# constants
MIN_DUP = 4
MIN_FULL_CTG_DUP_FRACTION = 0.40

class ITDPredictorCls(BasePredictorCls): #{
  def __init__(self, options, log_info=dict()): #{
    BasePredictorCls.__init__(self, 'itd', 'itd',
      'internal tandem duplication', options, log_info=log_info)
    self.good_topologies = set([
      'gap-tandem-duplication',
      'gap-nontandem-duplication',
    ])
    if (options.allow_non_gap_itds): #{
      self.good_topologies.update([
        'end-duplication',
        'intrachr-non-colinear',
        'junction-duplication',
      ])
    #} end if
    #self.store_seq = True
    full_ctg_dup_path = GetFilePath(options.output_dir, options.lib,
      "%s.full_ctg_dup" % self.ext)
    self.full_ctg_dup_file = FileBoxCls(full_ctg_dup_path, "w", "cannot "
      "create full contig duplication %s output file" % self.description)
    edge_gap_path = GetFilePath(options.output_dir, options.lib,
      "%s.edge_gap" % self.ext)
    self.edge_gap_file = FileBoxCls(edge_gap_path, "w", "cannot create "
      "edge-gap duplication %s output file" % self.description)
  #} end def

  def TestGroup(self, group): #{
    DebugMsg(self, "Group %s. Topologies: %s" %
      (group.id, ",".join(group.topologies)))
    # check alignment topology
    if (0 == len(self.good_topologies.intersection(group.topologies))): #{
      DebugMsg(self, "wrong alignment topology")
      return False
    #} end if
    return True
  #} end def

  def TestMember(self, member): #{
    DebugMsg(self, "  Member %s. Topology: %s" %
      (member.candidate_id, member.topology))
    # ignore members with the wrong topologies
    if (member.topology not in self.good_topologies): #{
      DebugMsg(self, "  %s: Wrong alignment topology: %s" %
        (member.IDString(), member.topology))
      return False
    #} end if
    # if the event matches two exon boundaries, it is not an ITD
    if (2 == member.meta_fields['exon_bounds']): #{
      DebugMsg(self, "  %s: Too many exon boundaries" % member.IDString())
      return False
    #} end if
    DebugMsg(self, "  valid number of exon boundaries")
    # if the member has a gap topology
    if (member.gap): #{
      # if the event coordinates are not internal to the contig
      if (not member.GapIsInternal()): #{
        DebugMsg(self, "  gap is not internal!")
        # then the member might not be an ITD
        if (self.options.require_internal_gaps): #{
          return False
        #} end if
        # check the edge gap fraction
        if (None != self.options.min_edge_gap_fraction): #{
          align_len    = float(member.align_info_B.ContigSpan())
          gap_len      = float(member.GapSpan())
          gap_fraction = align_len / gap_len
          if (1 < gap_fraction): #{
            LogMsg(self, "  WARNING: gapped candidate with aligned length "
               "greater than gap length!\n%s" % member.DataString())
          #} end if
          # if not enough of the gap is aligned, it is not an ITD
          if (self.options.min_edge_gap_fraction > gap_fraction): #{
            DebugMsg(self, "  not enough gap realigned!")
            return False
          #} end if
          DebugMsg(self, "  sufficient gap realignment")
        #} end if
      #} end if
      DebugMsg(self, "  member with internal gap")
    #} end if
    # if the member overlaps an annotated repeat region,
    # it might not be an ITD
    if (self.options.filter_itd_repeats and (member.OverlapsRepeat())): #{
      DebugMsg(self, "  member overlaps repeat!")
      return False
    #} end if
    # if the gene set is not empty, then the member might be an ITD
    return self.ITDGeneCheck(member)
    #if (member.ITDGeneCheck(exclude_non_coding=True)): #{
    #  self.bio_type_msg += ", %s" % member.bio_type_msg
    #  return True
    #} end if
  #} end def

  def ITDGeneCheck(self, member): #{
    if (member.gap): #{
      if (member.OverlapsGene("breakpoint_A", exclude_UTRs=True,
          exclude_non_coding=self.options.exclude_non_coding)):
        DebugMsg(self, "  overlapping a gene")
        return True
      #} end if
      DebugMsg(self, "  not overlapping any gene")
    else:
      # check that neither gene set is empty
      if (not member.OverlapsGene("breakpoint_A", exclude_UTRs=True,
          exclude_non_coding=self.options.exclude_non_coding) or
          not member.OverlapsGene("breakpoint_B", exclude_UTRs=True,
          exclude_non_coding=self.options.exclude_non_coding)): #{
        DebugMsg(self, "  empty gene set")
        return False
      #} end if
      # check that the gene sets overlap
      if (member.GeneSetsOverlap()): #{
        DebugMsg(self, "  with overlapping gene sets")
        return True
      #} end if
      DebugMsg(self, "  with disjoint gene sets")
    #} end if
    DebugMsg(self, "  failed internal tandem duplication gene check")
    return False
  #} end def

  def OutputEvent(self, event): #{
    # write edge-gap events to separate files
    if (self.write_special): #{
      for member in event.members: #{
        if (member.gap and not member.GapIsInternal()): #{
          if (MIN_FULL_CTG_DUP_FRACTION <
              member.align_info_B.align_fraction): #{
            DebugMsg(self, "Full contig duplication event")
            # write the event to the full contig duplication output file
            self.full_ctg_dup_file.Write(event.FullDataString())
          else:
            DebugMsg(self, "Edge-gap duplication event")
            # write the event to the edge-gap output file
            self.edge_gap_file.Write(event.FullDataString())
          #} end if
          return
        #} end if
      #} end for
    #} end if
    BasePredictorCls.OutputEvent(self, event)
  #} end def

  #def ReprocessMember(self, member, contig_alignment, contig_seq): #{
  def ReprocessMember(self, member, realigned_contig): #{
    contig_alignment = realigned_contig.best_align
    #if (None == contig_alignment): #{
    #  LogMsg(self, "  No alignments for contig %s" % member.contig_info.id)
    #  return True
    #} end if
    DebugMsg(self, "  Realign:%s" % contig_alignment.ToString())
    if (contig_alignment.perfect): #{
      DebugMsg(self, "  False positive: perfect alignment (%i)" %
        contig_alignment.match)
      return False
    #} end if
    # check any "gap" before the alignment
    gap_span = (contig_alignment.ctg_start - member.StartOfContig())
    if (MIN_DUP <= gap_span): #{
      gap_start = member.StartOfContig()-1
      gap_end   = contig_alignment.ctg_start-1
      gap_seq   = realigned_contig.sequence[gap_start:gap_end]
      if (not IsHomopolymerSequence(gap_seq)): #{
        target_start = contig_alignment.ctg_start-1
        target_end   = target_start + gap_span
        target_seq   = realigned_contig.sequence[target_start:target_end]
        best_match = MaxMatchLength(gap_seq, target_seq)
        min_match = max(MIN_DUP,
          int(gap_span*self.options.min_edge_gap_fraction))
        DebugMsg(self, "BEFORE\nGap:  %s\nTarg: %s\nMatch: %i Min: %i" %
          (gap_seq, target_seq, best_match, min_match))
        if (min_match <= best_match): #{
          DebugMsg(self, "  Good itd prediction: duplicated gap!")
          return True
        #} end if
      #} end if
    #} end if
    # check any "gap" after the alignment
    gap_span = (member.EndOfContig() - contig_alignment.ctg_end)
    if (MIN_DUP <= gap_span): #{
      gap_start = contig_alignment.ctg_end
      gap_end   = member.EndOfContig()
      gap_seq   = realigned_contig.sequence[gap_start:gap_end]
      if (not IsHomopolymerSequence(gap_seq)): #{
        target_end   = contig_alignment.ctg_end
        target_start = target_end - gap_span
        target_seq   = realigned_contig.sequence[target_start:target_end]
        best_match = MaxMatchLength(gap_seq, target_seq)
        min_match = max(MIN_DUP,
          int(gap_span*self.options.min_edge_gap_fraction))
        DebugMsg(self, "AFTER\nGap:  %s\nTarg: %s\nMatch: %i Min: %i" %
          (gap_seq, target_seq, best_match, min_match))
        if (min_match <= best_match): #{
          DebugMsg(self, "  Good itd prediction: duplicated gap!")
          return True
        #} end if
      #} end if
    #} end if
    # check any gaps within the alignment
    if (0 < contig_alignment.num_query_gaps): #{
      prev_right = contig_alignment.query_blocks[0][1]
      for block in contig_alignment.query_blocks[1:]: #{
        DebugMsg(self, "P:%i B:%i-%i" % (prev_right, block[0], block[1]))
        gap_span = (block[0] - prev_right) - 1
        if (MIN_DUP <= gap_span): #{
          DebugMsg(self, "Gap: %i-%i" % (prev_right+1, block[0]-1))
          gap_start = prev_right
          gap_end   = block[0]-1
          gap_seq   = realigned_contig.sequence[gap_start:gap_end]
          if (IsHomopolymerSequence(gap_seq)): #{
            DebugMsg(self, "Skipping homopolymer gap: %s" % gap_seq)
            prev_right = block[1]
            continue
          #} end if
          # check before the gap
          target_end   = gap_start
          target_start = target_end - gap_span
          target_seq   = realigned_contig.sequence[target_start:target_end]
          best_match = MaxMatchLength(gap_seq, target_seq)
          DebugMsg(self, "INTERNAL BEFORE\nGap:  %s\nTarg: %s\nMatch: %i "
            "Min: %i" % (gap_seq, target_seq, best_match, MIN_DUP))
          if (MIN_DUP <= best_match): #{
            DebugMsg(self, "  Good itd prediction: duplicated gap!")
            return True
          #} end if
          # check after the gap
          target_start = block[0]-1
          target_end   = target_start + gap_span
          target_seq   = realigned_contig.sequence[target_start:target_end]
          best_match = MaxMatchLength(gap_seq, target_seq)
          DebugMsg(self, "INTERNAL AFTER\nGap:  %s\nTarg: %s\nMatch: %i" %
            (gap_seq, target_seq, best_match))
          if (MIN_DUP <= best_match): #{
            DebugMsg(self, "  Good itd prediction: duplicated gap!")
            return True
          #} end if
        #} end if
        prev_right = block[1]
      #} end for
    #} end for
    # KLUDGE
    if (not member.gap): #{
      LogMsg(self, "  WARNING: automatically passing non-gap ITD prediction: "
        "%s" % realigned_contig.id)
      return True
    #} end if
    DebugMsg(self, "  No gaps have flanking duplications")
    return False
  #} end def
#} end class
