#! /usr/bin/env python

# TODO:
# Output lib_count in "across all libraries" genes list
# implement --filter-members option
#   (add this to quick_filter and bubble_up too)
# merge "annotations" and "filter-gene-names" options

# DONE:

# import standard modules
from optparse import OptionParser
import os, re, sys, time

# import custom modules
from utils.log import GetLogPath, CloseLogFile
from utils.error import MyError
from utils.general import (SetupMainClass, TimeSpent, WriteCommand,
  ShouldChromUseChr, RunOverlapCode, CleanExonName, CheckConfigCommands)
from utils.messages import ErrMsg, LogMsg, DebugMsg, ExtremeDebugMsg
from utils.files_paths import (FileBoxCls, CheckFilePath, CheckDirPath,
                               GetOutDir, EnsureAbsPath)
from parsers.candidate_group_parser import CandidateGroupParserCls
from parsers.candidate_group import CandidateGroupError
from fusion_predictor import FusionPredictorCls
from ptd_predictor import PTDPredictorCls
from itd_predictor import ITDPredictorCls
from realigner import RealignerCls

# CONSTANTS
ES_SUCCESS = 0
ES_RUN_ERR = 1
ES_OPT_ERR = 2
ES_PATH_ERR = 3

class EventPredictionCls: #{
  def __init__(self, options): #{
    SetupMainClass(self, options)
    if (self.options.realign): #{
      CheckConfigCommands(self, "blat")
    #} end if
    self.predictors = dict()
    if (self.options.predict_fusions): #{
      predictor = FusionPredictorCls(options, log_info=self.log_info)
      self.predictors[predictor.key] = predictor
    #} end if
    if (self.options.predict_ptds): #{
      predictor = PTDPredictorCls(options, log_info=self.log_info)
      self.predictors[predictor.key] = predictor
    #} end if
    if (self.options.predict_itds): #{
      predictor = ITDPredictorCls(options, log_info=self.log_info)
      self.predictors[predictor.key] = predictor
    #} end if
    self.use_chr = False
    #self.postpone_gene_check = False
  #} end def

  def __del__(self): #{
    CloseLogFile(self)
  #} end def

  def PredictEvents(self): #{
    LogMsg(self, "Predicting events...")
    start = time.time()
    # get the reference gene names, if a path is given
    #self.ref_gene_names = GetGeneNamesFromFile(self.options.gene_names_path,
    #  self.log_info)
    group_parser = CandidateGroupParserCls(self.options.groups_path)
    # recheck breakpoint exons
    self.RecheckBreakpointExons(group_parser)
    realigner = None
    if (self.options.realign): #{
      realigner = RealignerCls(self.options, self.log_info)
    #} end if
    # potential_events[bio_type][group_id] = event and gene sets object
    #potential_events = dict([(predictor.key, dict()) for
    #  predictor in self.predictors])
    LogMsg(self, "Processing candidate groups...")
    process_start = time.time()
    for group in group_parser: #{
      # get the breakpoint exons for the group
      self.GetBreakpointExons(group)
      # check whether the event is any biologically typed event
      #self.CheckEvent(group, output_files, lib_info.lib_name, potential_events)
      # attempt to predict events of each specified type
      # from the current candidate group
      for predictor in self.predictors.itervalues(): #{
        good_members = list()
        if (predictor.ProcessGroup(group, good_members) and
            None != realigner): #{
          #realigner.UpdateContigs(group, good_members, predictor.store_seq)
          realigner.UpdateContigs(group, good_members, predictor.key)
        #} end if
      #} end for
    #} end for
    LogMsg(self, "Time spent processing candidate groups: %s" %
      TimeSpent(process_start))
    #if ('event_coords' in output_files): #{
    #  output_files['event_coords'].Close()
    #  self.RecheckExonOverlap(output_files, potential_events, lib_info.lib_name)
    #} end if
    if (None != realigner and 0 < len(realigner.contigs)): #{
      realigner.RealignContigs()
      LogMsg(self, "Before realignment:")
      for predictor in self.predictors.itervalues(): #{
        LogMsg(self, "  Number of %s predictions: %i" %
          (predictor.description, predictor.num_predictions))
        if (0 == predictor.num_predictions): #{
          continue
        #} end if
        predictor.ReprocessPredictions(realigner.contigs)
        #predictor.ReprocessPredictions(realigner.contigs,
        #  realigner.contig_seqs)
      #} end for
      LogMsg(self, "%s\nAfter realignment:" % ("-"*40))
    #} end if
    for predictor in self.predictors.itervalues(): #{
      LogMsg(self, "Number of %s predictions: %i" %
        (predictor.description, predictor.num_predictions))
    #} end for
    LogMsg(self, "Time spent predicting events: %s" % TimeSpent(start))
  #} end def

  #def CreateOutputFiles(self, input_path): #{
  #  input_file_name = os.path.basename(input_path)
  #  input_root = os.path.splitext(input_file_name)[0]
  #  output_files = dict()
  #  # setup the coordinates file for rechecking exon overlaps
  #  self.SetupEventCoordsFile(input_root, output_files)
  #  return output_files
  #} end def

  def RecheckBreakpointExons(self, group_parser): #{
    if (None == self.options.breakpoint_exons): #{
      self.overlaps_file = None
      return
    #} end if
    group_coords_file = self.CreateGroupCoordsFile()
    for group in group_parser: #{
      ExtremeDebugMsg(self, "Writing coordinates for group %i" % group.id)
      self.WriteGroupCoords(group, group_coords_file)
    #} end for
    group_parser.Close()
    group_coords_file.Close()
    self.CreateOverlapsFile(group_coords_file.path)
  #} end def

  #def SetupEventCoordsFile(self, input_root, output_files): #{
  def CreateGroupCoordsFile(self): #{
    # check whether to use "chr" in chromosome names in coordinates file
    self.use_chr = ShouldChromUseChr(1, self.options.breakpoint_exons,
      "exon coordinates", self.log_info)
    # get the path for, and open the group coordinates file
    input_file_name = os.path.basename(self.options.groups_path)
    input_root = os.path.splitext(input_file_name)[0]
    group_coords_path = os.path.join(self.options.output_dir,
      "%s.coords" % input_root)
    #output_files['event_coords'] = FileBoxCls(group_coords_path, "w",
    group_coords_file = FileBoxCls(group_coords_path, "w",
      "cannot create event coordinates file")
    #self.postpone_gene_check = True
    return group_coords_file
  #} end def

  #def WriteEventCoords(self, event, group_coords_file): #{
  def WriteGroupCoords(self, event, group_coords_file): #{
    for member in event.members: #{
      if (member.gap): #{
        # write gap event coordinates
        self.WriteGapGroupCoords(member, group_coords_file)
      else:
        # write split event coordinates
        self.WriteSplitGroupCoords(member, group_coords_file)
      #} end if
    #} end for
  #} end def

  def WriteGapGroupCoords(self, member, group_coords_file): #{
    gap_coords = GroupCoordsCls(
      member.align_info_B.chrom,
      min(member.align_info_B.genome_start, member.align_info_B.genome_end),
      max(member.align_info_B.genome_start, member.align_info_B.genome_end),
      "%sA" % member.IDString(),
      self.use_chr
    )
    group_coords_file.WriteLine("%s" % gap_coords.ToString())
  #} end def

  def WriteSplitGroupCoords(self, member, group_coords_file): #{
    split_coords_A = GroupCoordsCls(
      member.align_info_A.chrom,
      member.align_info_A.genome_end - self.options.event_buffer,
      member.align_info_A.genome_end + self.options.event_buffer,
      "%sA" % member.IDString(),
      self.use_chr
    )
    group_coords_file.WriteLine("%s" % split_coords_A.ToString())
    split_coords_B = GroupCoordsCls(
      member.align_info_B.chrom,
      member.align_info_B.genome_start - self.options.event_buffer,
      member.align_info_B.genome_start + self.options.event_buffer,
      "%sB" % member.IDString(),
      self.use_chr
    )
    group_coords_file.WriteLine("%s" % split_coords_B.ToString())
  #} end def

  #def RecheckExonOverlap(self, output_files, potential_events, lib_name): #{
  #  LogMsg(self, "Rechecking exon overlap...")
  #  start = time.time()
  #  # run overlap code
  #  overlaps_path = self.RunOverlapCode(output_files['group_coords'].path)
  #  try: #{
  #    # parse overlap code output
  #    self.ParseOverlapResults(overlaps_path, potential_events)
  #  except ACEventGroupError, e:
  #    raise EventPredictionError("error parsing overlap file: %s" % e)
  #  #} end try
  #  self.ProcessPotentialEvents(potential_events, output_files, lib_name)
  #  LogMsg(self, "Time spent rechecking exon overlaps: %s" % TimeSpent(start))
  #} end def

  def CreateOverlapsFile(self, group_coords_path): #{
    LogMsg(self, "Running overlap code...")
    if (hasattr(self, "log_file") and None != self.log_file): #{
      self.log_file.Flush()
    #} end if
    start = time.time()
    overlaps_path = group_coords_path.replace(".coords", ".olap")
    RunOverlapCode(self.options.breakpoint_exons,
      group_coords_path, overlaps_path, dpt=self.options.dpt)
    LogMsg(self, "Time spent running overlap code: %s" % TimeSpent(start))
    self.overlaps_file = FileBoxCls(overlaps_path, "r",
      "cannot read exon/group overlaps file")
    #self.GetNextExonOverlap()
  #} end def

  def GetBreakpointExons(self, group): #{
    if (not hasattr(self, "overlaps_file") or None == self.overlaps_file): #{
      return
    #} end if
    ExtremeDebugMsg(self, "Getting breakpoint exons for group %i" % group.id)
    # clear any previous breakpoint genes
    group.ClearBPGenes()
    if (not hasattr(self, "curr_overlap")): #{
      self.curr_overlap = None
    #} end if
    # skip overlaps for groups that come before the current group
    while (None == self.curr_overlap or
        self.curr_overlap.group_id < group.id): #{
      try: #{
        self.GetNextExonOverlap()
      except StopIteration:
        return
      #} end try
    #} end while
    # create a dictionary of the members of the current group
    members_dict = dict()
    for member in group.members: #{
      members_dict[member.candidate_id] = member
    #} end for
    # get all overlaps for the current group
    while (self.curr_overlap.group_id == group.id): #{
      if (self.curr_overlap.member_id in members_dict): #{
        self.AddBreakPointGene(members_dict[self.curr_overlap.member_id])
      #} end if
      try: #{
        self.GetNextExonOverlap()
      except StopIteration:
        return
      #} end try
    #} end while
  #} end def

  def GetNextExonOverlap(self): #{
    if (not hasattr(self, "overlaps_file") or None == self.overlaps_file): #{
      ExtremeDebugMsg(self, "Setting current overlap to \"None\".")
      self.curr_overlap = None
      return
    #} end if
    overlap = ExonOverlapCls(self.overlaps_file.next())
    self.curr_overlap = overlap
    ExtremeDebugMsg(self, "Current overlap = G%i%s r%s exons: %s" %
      (overlap.group_id, overlap.member_id, overlap.region_id,
      ",".join(overlap.exons)))
  #} end def

  def AddBreakPointGene(self, member): #{
    if (None == self.curr_overlap): #{
      return
    #} end if
    if (self.curr_overlap.group_id != member.group_id): #{
      raise EventPredictionError("Group ID: %i does not match overlap ID: %i" %
        (member.group_id, self.curr_overlap.group_id))
    #} end if
    if (self.curr_overlap.member_id != member.candidate_id): #{
      raise EventPredictionError("Candidate ID: %s " % member.candidate_id +
        "does not match overlap ID: %s" % self.curr_overlap.member_id)
    #} end if
    member.AddGenes("breakpoint_%s" % self.curr_overlap.region_id,
      self.curr_overlap.exons)
  #} end def

  #def CheckEvent(self, event, output_files, lib_name, potential_events): #{
  #  # clear the event's messages
  #  event.bio_type_msg = None
  #  # filter the gene names
  #  #event.FilterGeneNames(self.ref_gene_names)
  #  DebugMsg(self, "Checking event:\n%s" % event.FullDataString())
  #  write_event_coords = False
  #  # check whether the event is a full exon duplication (PTD)
  #  if (self.options.predict_ptds): #{
  #    if (event.IsPTD(self.postpone_gene_check)): #{
  #      DebugMsg(self, "  PTD found!")
  #      if ('event_coords' in output_files): #{
  #        write_event_coords = True
  #        potential_events['ptds'][event.id] = EventAndGeneSetsCls(event)
  #      else:
  #        self.ProcessBiologicalEvent('ptds', lib_name, event, output_files)
  #      #} end if
  #    else:
  #      DebugMsg(self, "  NOT PTD: %s" % event.bio_type_msg)
  #    #} end if
  #  #} end if
  #  # check whether the event is a partial exon duplication (ITD)
  #  if (self.options.predict_itds): #{
  #    if (event.IsITD
  #          (self.postpone_gene_check,
  #           self.options.filter_itd_repeats,
  #           self.options.require_internal_gaps,
  #           self.options.min_edge_gap_fraction,
  #           self.options.allow_non_gap_itds)
  #    ):
  #      DebugMsg(self, "  ITD found!")
  #      if ('event_coords' in output_files): #{
  #        write_event_coords = True
  #        potential_events['itds'][event.id] = EventAndGeneSetsCls(event)
  #      else:
  #        self.ProcessBiologicalEvent('itds', lib_name, event, output_files)
  #      #} end if
  #    else:
  #      DebugMsg(self, "  NOT ITD: %s" % event.bio_type_msg)
  #    #} end if
  #  #} end if
  #  # check whether the event is a fusion
  #  warnings = list()
  #  if (self.options.predict_fusions): #{
  #    if (event.IsFusion
  #          (self.gene_dirs, warnings, self.postpone_gene_check)):
  #      if (0 < len(warnings)): #{
  #        LogMsg(self, "Warning: %s" % "\nWarning: ".join(warnings))
  #      #} end if
  #      DebugMsg(self, "  Fusion found!")
  #      if ('event_coords' in output_files): #{
  #        write_event_coords = True
  #        potential_events['fusions'][event.id] = EventAndGeneSetsCls(event)
  #      else:
  #        self.ProcessBiologicalEvent('fusions', lib_name, event, output_files)
  #      #} end if
  #    else:
  #      if (0 < len(warnings)): #{
  #        LogMsg(self, "Warning: %s" % "\nWarning: ".join(warnings))
  #      #} end if
  #      DebugMsg(self, "  NOT fusion: %s" % event.bio_type_msg)
  #    #} end if
  #  #} end if
  #  #if (write_event_coords): #{
  #  #  self.WriteEventCoords(event, output_files['event_coords'])
  #  #} end if
  #} end def

  #def ParseOverlapResults(self, overlaps_path, potential_events): #{
  #  LogMsg(self, "Parsing overlaps file...")
  #  start = time.time()
  #  # open the overlaps file
  #  overlaps_file = OpenFile(overlaps_path, "r", "cannot read overlaps file")
  #  # iterate through the overlaps file
  #  for overlap_line in overlaps_file: #{
  #    overlap_line = CleanLine(overlap_line)
  #    # skip blank lines
  #    if ("" == overlap_line): #{
  #      continue
  #    #} end if
  #    # parse the overlap line
  #    (group_id, member_id,
  #     region_id, new_exons) = self.ParseOverlapLine(overlap_line)
  #    # add the exon to the appropriate gene sets
  #    for predictor in self.predictors: #{
  #      if (group_id in potential_events[predictor.key]): #{
  #        DebugMsg(self, "Adding exons to group %i: %s" %
  #          (group_id, ", ".join(sorted(new_exons))))
  #        potential_events[predictor.key][group_id].AddExons(member_id,
  #          region_id, new_exons)
  #      #} end if
  #    #} end for
  #  #} end for
  #  overlaps_file.close()
  #  LogMsg(self, "Time spent parsing overlaps file: %s" % TimeSpent(start))
  #} end def

  #def ProcessPotentialEvents(self, potential_events, output_files, lib_name): #{
  #  DebugMsg(self, "Processing potential events after "
  #    "rechecking exon overlap...")
  #  for predictor in self.predictors: #{
  #    DebugMsg(self, "Potential %s events..." % predictor.description)
  #    for group_id in sorted(potential_events[predictor.key].keys()): #{
  #      DebugMsg(self, "GRPNUM: %i" % group_id)
  #      overlaps_genes = False
  #      event_and_gene_sets = potential_events[predictor.key][group_id]
  #      for member in event_and_gene_sets.event.members: #{
  #        member.breakpoint_genes_A = sorted(
  #          event_and_gene_sets.GetExons(member.member_id, "A"))
  #        member.breakpoint_genes_B = sorted(
  #          event_and_gene_sets.GetExons(member.member_id, "B"))
  #        if (0 < len(member.breakpoint_genes_A) and
  #            (0 < len(member.breakpoint_genes_B) or member.gap)):
  #          msg = ("Event %i breakpoint overlaps genes: %s" %
  #            (group_id, ",".join(member.breakpoint_genes_A)))
  #          if (0 < len(member.genes_B)): #{
  #            msg += ";%s" % ",".join(member.breakpoint_genes_B)
  #          #} end if
  #          DebugMsg(self, msg)
  #          overlaps_genes = True
  #        #} end if
  #      #} end for
  #      if (overlaps_genes): #{
  #        event = event_and_gene_sets.event
  #        if (event.BioTypeGeneCheck(predictor.key)): #{
  #          DebugMsg(self, "Event passed gene check: %i: %s" %
  #            (event.id, event.bio_type_msg))
  #          self.ProcessBiologicalEvent(predictor.key,
  #            lib_name, event, output_files)
  #        else:
  #          DebugMsg(self, "Event failed gene check: %i: %s" %
  #            (event.id, event.bio_type_msg))
  #        #} end if
  #      elif (self.debug):
  #        LogMsg(self, "Event %i breakpoint does not overlap genes" % group_id)
  #      #} end if
  #    #} end for
  #  #} end for
  #} end def
#} end class

class GroupCoordsCls: #{
  def __init__(self, chrom, left, right, id, use_chr): #{
    self.chrom = chrom
    self.left  = left
    self.right = right
    self.id    = id
    self.use_chr = use_chr
  #} end def

  def ToString(self): #{
    if (self.use_chr): #{
      chromosome = "chr%s" % self.chrom
    else:
      chromosome = self.chrom
    #} end if
    data_list = list([
      chromosome,
      "%i" % self.left,
      "%i" % self.right,
      self.id,
    ])
    return " ".join(data_list)
  #} end def
#} end class

class ExonOverlapCls: #{
  def __init__(self, overlap_line, log_info=dict()): #{
    self.log_info = log_info
    DebugMsg(self, "OVERLAP: %s" % overlap_line)
    # split the line into fields
    (chrom, left, right, id_str, exon_str) = overlap_line.split(" ", 4)
    # get the group, member, and region IDs
    id_patt = r"^(?P<group>\d+)(?P<member>[a-z]+)\)?(?P<region>[AB])$"
    id_match = re.search(id_patt, id_str)
    if (None == id_match): #{
      raise EventPredictionError("could not get ids from overlap line: "
        "\"%s\"\n" % id_str + "  %s" % overlap_line)
    #} end if
    self.group_id = int(id_match.group('group'))
    self.member_id = id_match.group('member')
    self.region_id = id_match.group('region')
    # get the exons overlapped
    exon_str = exon_str.replace(" ", "_")
    self.exons = set([CleanExonName(exon) for
      exon in exon_str.lstrip("^").split("^")])
  #} end def
#} end class

#### EXCEPTION CLASSES ####
class EventPredictionError(MyError): #{
  pass
#} end class

def SetupOptionsParser(): #{
  description_string = ("Predict events of specific types")
  args = [ "LIB_NAME", "GROUPS_FILE", ]
  usage_string = "%prog " + " ".join(args) + " [ OPTIONS ]"
  parser = OptionParser(description=description_string,
                        version="%prog 0.1.2",
                        usage=usage_string)
  parser.num_args = len(args)
  #parser.add_option("--input-is-list",
  #                  action="store_true", dest="list_input",
  #                  help="GROUPS_FILE is a list of results files for different "
  #                       "libraries [default].")
  #parser.add_option("--input-is-results",
  #                  action="store_false", dest="list_input",
  #                  help="GROUPS_FILE is a single results file.")
  parser.add_option("--with-gene-directions",
                    action="store_true", dest="use_gene_directions",
                    help="Only predict a fusion event if the transcription "
                         "direction of the genes involved is continuous "
                         "across the contig. [default]")
  parser.add_option("--without-gene-directions",
                    action="store_false", dest="use_gene_directions",
                    help="Do not consider transcription direction of genes "
                         "involved.")
  parser.add_option("--transcript-annotations",
                    metavar="PATH",
                    help="PATH is the path to a file containing gene "
                         "transcript annotations. Used for gene directions.")
                         #"and ITD realignment.")
  parser.add_option("--ignore-conflicts",
                    action="store_true",
                    help="When a gene has conflicting directions in the "
                         "annotations file, just ignore that gene, rather "
                         "than halting the script.")
  parser.add_option("--use-conflicts",
                    action="store_true",
                    help="When a gene has conflicting directions in the "
                         "annotations file, believe that the gene really "
                         "does go both ways.")
  #parser.add_option("--filter-gene-names",
  #                  dest="gene_names_path", metavar="FILE",
  #                  help="Use only gene names found in FILE (should be "
  #                       "tab-separated file, with gene name in column 13).")
  parser.add_option("--get-breakpoint-exons",
                    metavar="FILE", dest="breakpoint_exons",
                    help="Use the exon coordinates in FILE to get the exon "
                         "overlapped by the breakpoint in each event (used "
                         "for ITDs).")
  parser.add_option("--event-buffer",
                    type="int", metavar="N",
                    help="For split events, check whether the breakpoint "
                         "coordinates are within N bases of an exon. "
                         "[default: %default]")
  parser.add_option("--transcript-sequences",
                    metavar="PATH", dest="tran_seq_path",
                    help="The path to a file containing transcript sequences "
                         "to realign candidate contigs against, for avoiding "
                         "false positives due to problems with contig to "
                         "genome alignments.")
  parser.add_option("--contig-sequences",
                    metavar="PATH", dest="ctg_seq_path",
                    help="The path to a file containing candidate contig "
                         "sequences to realign to transcript sequences, for "
                         "avoiding false positives due to problems with "
                         "contig to genome alignments.")
  parser.add_option("--filter-ITD-repeats",
                    action="store_true", dest="filter_itd_repeats",
                    help="Do not predict an ITD event if the candidate "
                         "contig overlaps any annotated repeats.")
  parser.add_option("--require-internal-gaps",
                    action="store_true", dest="require_internal_gaps",
                    help="Require that the duplicated sequence is internal "
                         "to the contig for gapped ITD "
                         "event predictions. [default]")
  parser.add_option("--allow-edge-gaps",
                    action="store_false", dest="require_internal_gaps",
                    help="Allow the duplicated sequence to be at the very "
                         "edge of the contig for gapped partial exon "
                         "duplication predictions.")
  parser.add_option("--min-edge-gap-fraction",
                    type="float", metavar="F",
                    help="If the gap is at the very edge of a contig, "
                         "require that the fraction of the gap involved "
                         "in the duplication is at least F. This option "
                         "is ignored if the --require-internal-gaps option "
                         "is used. [default: %default]")
  parser.add_option("--allow-non-gap-ITDs",
                    action="store_true", dest="allow_non_gap_itds",
                    help="Look for ITDs in groups with junction duplication, "
                         "end duplication, and non-colinear topologies "
                         "as well as gap topologies.")
  parser.add_option("--exclude-non-coding",
                    action="store_true",
                    help="Do not predict ITD events in non-coding genes. "
                         "[default]")
  parser.add_option("--allow-non-coding",
                    action="store_false", dest="exclude_non_coding",
                    help="Report ITD events predicted in non-coding genes.")
  parser.add_option("--predict-fusions",
                    action="store_true", dest="predict_fusions",
                    help="Predict fusion events. [default]")
  parser.add_option("--no-fusions",
                    action="store_false", dest="predict_fusions",
                    help="Do no predict fusion events")
  parser.add_option("--predict-PTDs",
                    action="store_true", dest="predict_ptds",
                    help="Predict PTD events. [default]")
  parser.add_option("--no-PTDs",
                    action="store_false", dest="predict_ptds",
                    help="Do no predict PTD events")
  parser.add_option("--predict-ITDs",
                    action="store_true", dest="predict_itds",
                    help="Predict ITD events. [default]")
  parser.add_option("--no-ITDs",
                    action="store_false", dest="predict_itds",
                    help="Do no predict ITD events")
  parser.add_option("-p", "--pretty",
                    action="store_true",
                    help="Print more readable output file as well as "
                         "standard data file.")
  parser.add_option("--disable-profiling-timer",
                    action="store_true", dest="dpt",
                    help="Sometimes this script can hang when trying to spawn "
                         "child processes, due to the kernel's profiling "
                         "timer. Use this option to disable the profiling "
                         "timer if the script seems to be hanging.")
  parser.add_option("-f", "--force",
                    action="store_true",
                    help="Force prediction to take place, even if the output "
                         "directory already exists.")
  parser.add_option("-d", "--debug",
                    action="store_true",
                    help="Print debug information while the program runs.")
  parser.add_option("--extreme-debug",
                    action="store_true", dest="extreme_debug",
                    help="Print extremely in-depth debug information while "
                      "the program runs. Not recommended for large jobs.")
  parser.set_defaults(#list_input=True,
                      use_gene_directions=True,
                      ignore_conflicts=False,
                      use_conflicts=False,
                      event_buffer=5,
                      filter_itd_repeats=False,
                      require_internal_gaps=True,
                      min_edge_gap_fraction=0.80,
                      allow_non_gap_itds=False,
                      exclude_non_coding=True,
                      predict_fusions=True,
                      predict_ptds=True,
                      predict_itds=True,
                      pretty=False,
                      dpt=False,
                      force=False,
                      debug=False,
                      extreme_debug=False)
  return parser
#} end def

def CheckPaths(options): #{
  opts_good = True
  if (options.ignore_conflicts and options.use_conflicts): #{
    ErrMsg("Cannot use both \"--ignore-conflicts\" and \"--use-conflicts\" "
           "options at the same time. Please choose one or the other.")
    opts_good = False
  #} end if
  path_errors = list()
  options.realign = False
  if (None != options.tran_seq_path): #{
    CheckFilePath(options.tran_seq_path, "transcript sequences", path_errors)
    if(None == options.ctg_seq_path): #{
      ErrMsg("You must provide the path to a candidate contig sequences file "
        "when you provide a transcript sequences path.")
      opts_good = False
    #} end if
    options.realign = True
  #} end if
  if (None != options.ctg_seq_path): #{
    CheckFilePath(options.ctg_seq_path, "candidate contig sequences",
      path_errors)
    if(None == options.tran_seq_path): #{
      ErrMsg("You must provide the path to a transcript sequences file "
        "when you provide a candidate contig sequences path.")
      opts_good = False
    #} end if
  #} end if
  #if (options.list_input):
  #  input_path_err_msg = \
  #    "Cannot find libraries list file: \"%s\"." % options.groups_path
  #else:
  #  input_path_err_msg = \
  #    "Cannot find results file: \"%s\"." % options.groups_path
  #} end if
  #if (not os.path.isfile(options.groups_path)):
  #  path_errors.append(input_path_err_msg)
  #} end if
  CheckFilePath(options.groups_path, "groups", path_errors)
  # get the input directory
  (input_dir, file_name) = os.path.split(options.groups_path)
  #if (None != options.gene_names_path and
  #    not os.path.isfile(options.gene_names_path)):
  #  path_errors.append \
  #    ("Cannot find gene names annotation file: \"%s\"." %
  #     options.gene_names_path)
  #} end if
  # if checking the gene directions
  if (options.use_gene_directions): #{  or options.realign_itds):
    # ensure a gene annotations path was given
    if (None == options.transcript_annotations): #{
      path_errors.append("You must give the path to a gene transcript "
        "annotations file when using the --with-gene-directions option.")
        #"or --realign-ITDs option.")
    else:
      CheckFilePath(options.transcript_annotations,
        "gene transcript annotations", path_errors)
    #} end if
  #} end if
  if (None != options.breakpoint_exons): #{
    CheckFilePath(options.breakpoint_exons, "exon coordinate annotations",
      path_errors)
  #if (None != options.get_breakpoint_exons and
  #    not os.path.isfile(options.get_breakpoint_exons)):
  #  path_errors.append \
  #    ("Cannot find exon coordinates file: \"%s\"." %
  #     options.get_breakpoint_exons)
  #} end if
  #if (os.path.exists(options.output_dir)): #{
  #  path_errors.append \
  #    ("Output directory already exists: \"%s\"." % options.output_dir)
  #elif (0 == len(path_errors)):
  #  CreateDirectory(options.output_dir)
  #  if (not os.path.isdir(options.output_dir)): #{
  #    path_errors.append \
  #      ("Cannot create output directory: \"%s\"." % options.output_dir)
  #  #} end if
  #} end if
  options.output_dir = GetOutDir(input_dir, "predicted_events")
  if (opts_good and 0 == len(path_errors)): #{
    CheckDirPath(options.output_dir, "output", path_errors,
      create=True, replace=options.force)
    # get the log file name
    options.log_file_name = GetLogPath(options.groups_path,
      "prediction", options.output_dir)
    if (options.realign): #{
      options.realign_dir = os.path.join(options.output_dir, "realignment")
      CheckDirPath(options.realign_dir, "intermediate realignment",
        path_errors, create=True)
    #} end if
  #} end if
  if (0 < len(path_errors)): #{
    ErrMsg("Errors in input arguments:\n  %s" % "\n  ".join(path_errors))
  #} end if
  #if (None != options.gene_names_path and
  #    None != options.get_breakpoint_exons):
  #  ErrMsg("Cannot use both \"--filter-gene-names\" and "
  #          "\"--get-breakpoint-exons\" options at the same time. Please "
  #          "choose one or the other.")
  #  opts_good = False
  #} end if
  # the paths are good if there are no path errors and no conflicting options
  return (opts_good and 0 == len(path_errors))
#} end def

def Main(): #{
  # get options, args, and params
  parser = SetupOptionsParser()
  (options, args) = parser.parse_args()
  # if the right number of args was used
  if (parser.num_args == len(args)): #{
    options.lib         = args[0]
    options.groups_path = EnsureAbsPath(args[1])
    if (CheckPaths(options)): #{
      try: #{
        event_predictor = EventPredictionCls(options)
        WriteCommand(event_predictor, sys.argv)
        event_predictor.PredictEvents()
      except (MyError), e:
        ErrMsg("ERROR while getting biologically typed events in "
               "anomalous contigs:\n  %s" % e)
        return ES_RUN_ERR
      #} end try
    else:
      return ES_PATH_ERR
    #} end if
  else:
    parser.error("you must specify a library name (LIB_NAME); and the path "
                 "to a file containing BARNACLE group data for that library "
                 "(GROUPS_FILE).")
    return ES_OPT_ERR
  #} end if
  return ES_SUCCESS
#} end def

if __name__ == '__main__': #{
  exit_status = Main()
  if (ES_SUCCESS == exit_status): #{
    print "BIOLOGICAL TYPING SUCCESS"
  else:
    print "BIOLOGICAL TYPING FAIL"
  #} end if
  sys.exit(exit_status)
#} end if
