#! /usr/bin/env python

# import standard modules
from optparse import OptionParser
import os, re, sys, time

# import custom modules
from utils import submit
from utils.log import GetLogPath, CloseLogFile
from utils.error import MyError
from utils.general import (SetupMainClass, TimeSpent, WriteCommand, GetOptions,
                           CheckConfigCommands, GetCommand)
from utils.messages import ErrMsg, LogMsg, DebugMsg
from utils.files_paths import (FileBoxCls, CheckFilePath, CheckNewFilePath,
                               CheckDirPath, GetOutDir, EnsureAbsPath,
                               EnsureDirectoryExists, CleanLine)
from utils.subprocesses import (RunCommandFromString, SortFile)
from alignment_processing.alignment_functions import CalcOverlap
from parsers.candidate_group_parser import CandidateGroupParserCls

# CONSTANTS
ES_SUCCESS = 0
ES_RUN_ERR = 1
ES_OPT_ERR = 2
ES_PATH_ERR = 3

class R2CSubmitterCls: #{
  def __init__(self, options): #{
    SetupMainClass(self, options)
    if (None == self.options.cluster_head): #{
      local = True
    else:
      local = False
    #} end if
    CheckConfigCommands(self, "python", local)
    self.jobs = R2CJobCollectionCls(options, log_info=self.log_info)
    # group_to_jobs_map is a file
    self.group_to_jobs_map = None
    self.short_ctgs = set()
  #} end def

  def __del__(self): #{
    if (hasattr(self, "group_to_jobs_map") and
        None != self.group_to_jobs_map): #{
      self.group_to_jobs_map.Close()
    #} end if
    for job in self.jobs: #{
      job.Close()
    #} end for
    CloseLogFile(self)
  #} end def

  def Run(self): #{
    LogMsg(self, "Creating parallel read-to-contig jobs...")
    start = time.time()
    self.Setup()
    for group in self.groups_file: #{
      LogMsg(self, "Processing group %i..." % group.id)
      DebugMsg(self, group.FullDataString().rstrip())
      group_start = time.time()
      self.ProcessGroup(group)
      DebugMsg(self, "Time spent processing group: %s\n" %
        TimeSpent(group_start) + "--------------------------\n")
    #} end for
    if (None == self.options.cluster_head): #{
      self.RunJobsLocally()
    else:
      jobs_path = self.CreateJobsFile()
      self.SubmitJobs(jobs_path)
    #} end if
    if (self.options.no_short_ctgs and 0 < len(self.short_ctgs)): #{
      LogMsg(self, "WARNING: %i contigs shorter than the read-length have "
        "been ignored." % len(self.short_ctgs))
    #} end if
    LogMsg(self, "Time spent creating parallel read-to-contig "
      "jobs: %s" % TimeSpent(start))
  #} end def

  def Setup(self): #{
    self.groups_file = CandidateGroupParserCls(self.options.groups_path)
    fail_msg = "cannot create groups-to-jobs map file"
    groups_to_jobs_path = os.path.join(self.options.cluster_dir,
      "group2jobs_map")
    self.group_to_jobs_map = FileBoxCls(groups_to_jobs_path, "w", fail_msg)
  #} end def

  def ProcessGroup(self, group): #{
    job_indices = set()
    member_strings = list()
    # get member strings for group
    for member in group.members: #{
      member_strings.append(self.GetMemberString(member))
    #} end for
    # write member strings to appropriate jobs files
    for member_index in range(len(member_strings)): #{
      mate_string_list = \
        member_strings[:member_index] + member_strings[member_index+1:]
      if (0 == len(mate_string_list)): #{
        mates_string = "None"
      else:
        mates_string = ";".join(mate_string_list).replace(" ", ",")
      #} end if
      full_member_str = member_strings[member_index] + \
        " MATES:%s" % mates_string
      DebugMsg(self, "  MEMBER STRING: %s" % full_member_str)
      ctg_id = group.members[member_index].contig_info.id
      ctg_len = group.members[member_index].contig_info.length
      if (self.options.no_short_ctgs and
          self.options.read_length > ctg_len): #{
        self.short_ctgs.add(ctg_id)
        DebugMsg(self, "Not adding member to job: contig shorter "
          "than read-length.")
        continue
      #} end if
      job_index = self.jobs.ProcessMember(ctg_id, full_member_str)
      job_indices.add(str(job_index))
    #} end for
    jobs_string = "%i:%s" % (group.id, ",".join(job_indices))
    self.group_to_jobs_map.WriteLine(jobs_string)
  #} end def

  def GetMemberString(self, member): #{
    (left, right) = self.GetBufferedEventCoords(member)
    member_info = [
      ConvertContigID(member.contig_info.id),
      member.IDString(),
      "%i-%i" % (left, right),
    ]
    return " ".join(member_info)
  #} end def

  def GetBufferedEventCoords(self, member): #{
    if (member.gap): #{
      # if the gap is internal to the contig
      if (member.GapIsInternal()): #{
        left  = member.align_info_B.ctg_start
        right = member.align_info_B.ctg_end
      else:
        # figure out which coordinates to use
        realign_start = member.align_info_B.ctg_start
        realign_end   = member.align_info_B.ctg_end
        if ('dup_start' in member.meta_fields): #{
          (olapA, fractionA) = CalcOverlap(member.meta_fields['gap_start'],
            member.meta_fields['gap_end'], realign_start, realign_end)
          (olapB, fractionB) = CalcOverlap(member.meta_fields['gap_start'],
            member.meta_fields['gap_end'], member.meta_fields['dup_start'],
            member.meta_fields['dup_end'])
          ctg_str = "ctg:%i-%i(%i)" % (member.align_info_B.ctg_start,
            member.align_info_B.ctg_end, olapA)
          dup_str = "dup:%i-%i(%i)" % (member.meta_fields['dup_start'],
            member.meta_fields['dup_end'], olapB)
          DebugMsg(self, "Checking which coordinates to use for edge-gap\n"
            "gap:%i-%i %s %s" % (member.meta_fields['gap_start'],
             member.meta_fields['gap_end'], ctg_str, dup_str))
          if (olapB > olapA): #{
            realign_start = member.meta_fields['dup_start']
            realign_end   = member.meta_fields['dup_end']
            DebugMsg(self, "Using DUP coordinates")
          else:
            DebugMsg(self, "Using CTG coordinates")
          #} end if
        #} end if
        # if the gap is at the start of the contig
        if (member.StartOfContig() == member.meta_fields['gap_start']): #{
          left  = min(realign_end, member.meta_fields['gap_end'])
          right = max(realign_end, member.meta_fields['gap_end'])
        # if the gap is at the end of the contig
        else:
          left  = min(realign_start, member.meta_fields['gap_start'])
          right = max(realign_start, member.meta_fields['gap_start'])
        #} end if
      #} end if
    else:
      left  = min(member.align_info_A.ctg_end, member.align_info_B.ctg_start)
      right = max(member.align_info_A.ctg_end, member.align_info_B.ctg_start)
    #} end if
    # widen the region to satisfy the minimum overlap requirement and
    # ensure that the group does not extend past the edges of the contig
    min_left  = 2
    max_right = member.contig_info.length-1
    buffered_left  = max(left  - self.options.min_overlap, min_left)
    buffered_right = min(right + self.options.min_overlap, max_right)
    if ((buffered_right - buffered_left) > 1000): #{
      LogMsg(self, "WARNING: wide event coordinates: %i" %
        ((buffered_right - buffered_left) + 1))
    #} end if
    return (buffered_left, buffered_right)
  #} end def

  def CreateJobsFile(self): #{
    jobs_path = os.path.join(self.options.cluster_dir, "read2ctg_jobs")
    LogMsg(self, "Creating jobs file: %s" % jobs_path)
    fail_msg = "cannot create jobs file"
    jobs_file = FileBoxCls(jobs_path, "w", fail_msg)
    for job in self.jobs: #{
      job.Close()
      sort_start = time.time()
      job.SortFileByContig()
      LogMsg(self,
        "Time spent sorting job data: %s" % TimeSpent(sort_start))
      jobs_file.WriteLine(job.Command(local=False))
    #} end for
    jobs_file.Close()
    return jobs_file.path
  #} end def

  def RunJobsLocally(self): #{
    LogMsg(self, "Running jobs locally...")
    for job in self.jobs: #{
      job.Close()
      sort_start = time.time()
      job.SortFileByContig()
      LogMsg(self,
        "Time spent sorting job data: %s" % TimeSpent(sort_start))
      fail_msg = "cannot create job command file"
      job_file = FileBoxCls(job.CommandPath(), "w", fail_msg)
      job_file.WriteLine(job.Command(local=False))
      job_file.Close()
      fail_msg = "cannot create stdout file for job"
      out_file = {'path':job.StdOutPath(), 'mode':"w", 'fail_msg':fail_msg}
      fail_msg = "cannot create stderr file for job"
      err_file = {'path':job.StdErrPath(), 'mode':"w", 'fail_msg':fail_msg}
      job_command = job.Command(local=True)
      DebugMsg(self, job_command)
      return_code = RunCommandFromString(job_command,
        stdout=out_file, stderr=err_file, dpt=self.options.dpt)
      if (0 != return_code and not self.log_info['debug']): #{
        LogMsg(self, job_command)
      #} end if
      if (0 > return_code): #{
        raise R2CSubmitterError("Read-to-contig job command was terminated "
          "by signal %i" % return_code)
      elif (0 < return_code):
        raise R2CSubmitterError("Error running read-to-contig job "
          "command: %i" % return_code)
      #} end if
    #} end for
  #} end def

  def SubmitJobs(self, jobs_path): #{
    LogMsg(self, "Submitting read-to-contig jobs...")
    submit_options = GetOptions(submit)
    submit.SetSubmitOptions(submit_options, "%s-r2c" % self.options.lib,
      jobs_path, False, self.options)
    if (submit.CheckPaths(submit_options)): #{
      submitter = submit.JobSubmitterCls(submit_options,
        log_info=self.log_info)
      submitter.Submit()
    else:
      raise R2CSubmitterError("cannot resubmit jobs: "
        "errors in resubmit options")
    #} end if
  #} end def
#} end class

class R2CJobCollectionCls: #{
  def __init__(self, options, log_info=dict()): #{
    self.options  = options
    self.log_info = log_info
    # internally job list uses base 0, but external directories use base 1
    # job_lookup[ctg_id] = job_index
    self.job_lookup = dict()
    self.jobs = list()
  #} end def

  def __iter__(self): #{
    return self.jobs.__iter__()
  #} end def

  def ProcessMember(self, ctg_id, member_str): #{
    DebugMsg(self, "Processing member with contig: %s" % ctg_id)
    if (ctg_id not in self.job_lookup): #{
      self.AddNewContig(ctg_id)
    #} end if
    job_index = self.job_lookup[ctg_id]
    self.jobs[job_index].ProcessMember(member_str)
    # internally job list uses base 0, but external directories use base 1
    return self.jobs[job_index].index
  #} end def

  def AddNewContig(self, ctg_id): #{
    DebugMsg(self, "Adding new contig to job collection")
    if (0 == len(self.jobs) or self.jobs[-1].IsFull()): #{
      DebugMsg(self, "Adding new job to job collection, num jobs: %i" %
        (len(self.jobs)+1))
      new_job = R2CJobCls(len(self.jobs)+1, self.options,
        log_info=self.log_info)
      self.jobs.append(new_job)
    #} end if
    self.jobs[-1].AddNewContig(ctg_id)
    # internally job list uses base 0, but external directories use base 1
    self.job_lookup[ctg_id] = self.jobs[-1].index-1
  #} end def
#} end class

class R2CJobCls: #{
  def __init__(self, index, options, log_info=dict()): #{
    self.options  = options
    self.log_info = log_info
    self.index    = index
    self.contigs  = set()
    self.file     = None
  #} end def

  def IsFull(self): #{
    num_contigs = len(self.contigs)
    if (num_contigs < self.options.ctgs_per_job): #{
      return False
    #} end if
    return True
  #} end def

  def AddNewContig(self, ctg_id): #{
    self.contigs.add(ctg_id)
  #} end def

  def CreateJobDataFile(self): #{
    self.job_dir = os.path.join(self.options.cluster_dir,
      "job_%i" % self.index)
    EnsureDirectoryExists(self.job_dir)
    job_path = os.path.join(self.job_dir, "%s.%i.r2c.data" %
      (self.options.lib, self.index))
    fail_msg = "cannot create job data file"
    self.file = FileBoxCls(job_path, "w", fail_msg)
  #} end def

  def ProcessMember(self, member_str): #{
    if (None == self.file): #{
      self.CreateJobDataFile()
    #} end if
    self.file.WriteLine(member_str)
  #} end def

  def SortFileByContig(self): #{
    if (None == self.file): #{
      raise R2CSubmitterError \
        ("job data file not created for job %i" % self.index)
    #} end if
    self.Close()
    DebugMsg(self, "Sorting job %i data" % self.index)
    SortFile(self, self.file.path, 1, numeric=False)
  #} end def

  def Command(self, local=False): #{
    #python_path = RunCommandFromString("which python", stdout=STRING_OUT,
    #  dpt=self.options.dpt)[0]
    #python_path = CleanLine(python_path)
    #CheckFilePath(python_path, "python binary")
    python_path = GetCommand(self, "python", local)
    script_path = os.path.join(os.environ['BARNACLE_PATH'],
      "support", "read_to_contig", "calculate.py")
    CheckFilePath(script_path, "read-to-contig support script")
    command_list = ["time", python_path, script_path, self.options.lib,
      self.file.path, self.options.r2c_path, "--min-overlap",
      "%i" % self.options.min_overlap]
    if (self.log_info['debug']): #{
      command_list.append("--debug")
    #} end if
    command = " ".join(command_list)
    if (local): #{
      if ("r2c" in os.path.basename(self.file.path)): #{
        log_type = "calculate"
      else:
        log_type = "r2c.calculate"
      #} end if
      log_path = GetLogPath(self.file.path, log_type, options.output_dir)
      command += " --log-file %s" % log_path
      return command
    #} end if
    full_command_list = [
      "export BARNACLE_PATH=%s" % os.environ["BARNACLE_PATH"],
      "export PATH=$BARNACLE_PATH:$PATH",
      "export PYTHONPATH=.:$BARNACLE_PATH:$PYTHONPATH",
      command,
    ]
    full_command = ";".join(full_command_list)
    return full_command
  #} end def

  def CommandPath(self): #{
    return os.path.join(self.options.cluster_dir,
      "%s-r2c.%i.sh" % (self.options.lib, self.index))
  #} end def

  def StdOutPath(self): #{
    return "%s.o1" % self.CommandPath()
  #} end def

  def StdErrPath(self): #{
    return "%s.e1" % self.CommandPath()
  #} end def

  def Close(self): #{
    if (None != self.file): #{
      self.file.Close()
    #} end if
  #} end def
#} end def

def ConvertContigID(ctg_id): #{
  # replace ":", "+", and "-" with "_", removing commas and trailing "_"
  return re.sub(r"_$", "", re.sub(r"[:+-],?", "_", ctg_id))
#} end def

#### EXCEPTION CLASSES ####
class R2CSubmitterError(MyError): #{
  pass
#} end class

def SetupOptionsParser(): #{
  description_string = \
    "Create and submit parallel jobs for calculating read-to-contig support"
  args = [ "LIB", "GROUPS_FILE", "READ_TO_CONTIG_FILE", ]
  usage_string = "%prog " + " ".join(args) + " [ OPTIONS ]"
  parser = OptionParser(description=description_string,
                        version="%prog 0.1.2",
                        usage=usage_string)
  parser.num_args = len(args)
  parser.add_option("--min-overlap",
                    type="int", metavar="N",
                    help="Require that reads overlap breakpoints by at "
                         "least N bp. [default: %default]")
  parser.add_option("--ctgs-per-job",
                    type="int", metavar="N",
                    help="Process N contigs in each job. [default=%default]")
  parser.add_option("--no-short-ctgs",
                    action="store_true",
                    help="When creating read-to-conting support jobs, do not "
                      "create jobs for contigs shorter than the read length.")
  parser.add_option("--read-length",
                    type="int", metavar="N",
                    help="Reads are N bases long. This option is required "
                      "when using --no-short-ctgs.")
  parser.add_option("-c", "--cluster-head",
                    dest="cluster_head",
                    help="Cluster head node to submit to.")
  parser.add_option("--mem", "--memory",
                    help="The memory requirement of the jobs "
                         "[default:\"%default\"].")
  parser.add_option("--hostname",
                    help="The hostname(s) to submit to "
                         "[default:\"%default\"].")
  parser.add_option("--queue",
                    help="The queue(s) to submit to [default:\"%default\"].")
  parser.add_option("-w", "--wall-time",
                    metavar="H:MM:SS",
                    help="The maximum time to spend on the job.")
  parser.add_option("--email",
                    help="E-mail status updates on submitted jobs to the "
                         "given email address")
  parser.add_option("--disable-profiling-timer",
                    action="store_true", dest="dpt",
                    help="Sometimes this script can hang when trying to spawn "
                         "child processes, due to the kernel's profiling "
                         "timer. Use this option to disable the profiling "
                         "timer if the script seems to be hanging.")
  parser.add_option("-d", "--debug",
                    action="store_true",
                    help="Print debug information while the program runs.")
  parser.set_defaults(min_overlap=5,
                      ctgs_per_job=800,
                      no_short_ctgs=False,
                      mem="5G",
                      hostname="q*",
                      queue="all.q",
                      dpt=False,
                      debug=False)
  return parser
#} end def

def CheckPaths(options): #{
  path_errors = list()
  CheckFilePath(options.groups_path, "groups", path_errors)
  (groups_dir, groups_file_name) = os.path.split(options.groups_path)
  options.cluster_dir = os.path.join(os.path.dirname(groups_dir),
    "cluster_r2c")
  CheckFilePath(options.r2c_path, "read-to-contigs alignments", path_errors)
  if (0 == len(path_errors)): #{
    # attempt to create output directory, if it does not already exist
    CheckDirPath(options.cluster_dir, "cluster directory", path_errors,
      create=True)
  #} end if
  options.log_file_name = \
    GetLogPath(options.groups_path, "r2c.submit", options.cluster_dir)
  if (0 < len(path_errors)): #{
    ErrMsg("Errors in input arguments:\n  %s" % "\n  ".join(path_errors))
  #} end if
  opts_good = True
  if (options.no_short_ctgs and None == options.read_length): #{
    ErrMsg("You must provide the read-length when using the "
      "--no-short-ctgs option.")
    opts_good = False
  #} end if
  # the paths are good if there are no path errors and no conflicting options
  return (0 == len(path_errors) and opts_good)
#} end def

def Main(): #{
  # get options, args, and params
  parser = SetupOptionsParser()
  (options, args) = parser.parse_args()
  # if the right number of args was used
  if (parser.num_args == len(args)): #{
    options.lib           = args[0]
    options.groups_path   = EnsureAbsPath(args[1])
    options.r2c_path      = EnsureAbsPath(args[2])
    if (CheckPaths(options)): #{
      try: #{
        r2c_submitter = R2CSubmitterCls(options)
        WriteCommand(r2c_submitter, sys.argv)
        r2c_submitter.Run()
      except (MyError), e:
        ErrMsg("ERROR while creating parallel read-to-contig jobs:\n  %s" % e)
        return ES_RUN_ERR
      #} end try
    else:
      return ES_PATH_ERR
    #} end if
  else:
    parser.error("you must specify a library name (LIB); the path to a "
                 "file containing BARNACLE group data (GROUPS_FILE); "
                 "and the path to a bam-formatted file containing the read "
                 "to contig alignment results to use (READ_TO_CONTIG_FILE).")
    return ES_OPT_ERR
  #} end if
  return ES_SUCCESS
#} end def

if __name__ == '__main__': #{
  exit_status = Main()
  if (ES_SUCCESS == exit_status): #{
    print "R2C SUBMIT SUCCESS"
  else:
    print "R2C SUBMIT FAIL"
  #} end if
  sys.exit(exit_status)
#} end if
